from django.test.runner import DiscoverRunner

from behave_django.environment import BehaveHooksMixin
from behave_django.testcase import (BehaviorDrivenTestCase,
                                    ExistingDatabaseTestCase,
                                    DjangoSimpleTestCase)


class BehaviorDrivenTestRunner(DiscoverRunner, BehaveHooksMixin):
    """
    Test runner that uses the BehaviorDrivenTestCase
    """
    testcase_class = BehaviorDrivenTestCase


class ExistingDatabaseTestRunner(DiscoverRunner, BehaveHooksMixin):
    """
    Test runner that uses the ExistingDatabaseTestCase

    This test runner nullifies Django's test database setup methods. Using this
    test runner would make your tests run with the default configured database
    in settings.py.
    """
    testcase_class = ExistingDatabaseTestCase

    def setup_databases(self, **kwargs):
        pass

    def teardown_databases(self, old_config, **kwargs):
        pass


class SimpleTestRunner(DiscoverRunner, BehaveHooksMixin):
    """
    Test runner that uses DjangoSimpleTestCase with atomic
    transaction management and no support of web browser automation.
    """
    testcase_class = DjangoSimpleTestCase



from django.contrib.staticfiles.testing import StaticLiveServerTestCase
from django.test.testcases import TestCase


class BehaviorDrivenTestMixin(object):
    """
    Mixin to prevent the TestCase from executing its setup and teardown methods

    This mixin overrides the test case's _pre_setup and _post_teardown methods
    in order to prevent them from executing when the test case is instantiated.
    We do this to have total control over the test execution.
    """

    def _pre_setup(self, run=False):
        if run:
            super(BehaviorDrivenTestMixin, self)._pre_setup()

    def _post_teardown(self, run=False):
        if run:
            super(BehaviorDrivenTestMixin, self)._post_teardown()

    def runTest(self):
        pass


class BehaviorDrivenTestCase(BehaviorDrivenTestMixin,
                             StaticLiveServerTestCase):
    """
    Test case attached to the context during behave execution

    This test case prevents the regular tests from running.
    """


class ExistingDatabaseTestCase(BehaviorDrivenTestCase):
    """
    Test case used for the --use-existing-database setup

    This test case prevents fixtures from being loaded to the database in use.
    """

    def _fixture_setup(self):
        pass

    def _fixture_teardown(self):
        pass


class DjangoSimpleTestCase(BehaviorDrivenTestMixin, TestCase):
    """
    Test case attached to the context during behave execution

    This test case uses `transaction.atomic()` to achieve test isolation
    instead of flushing the entire database. As a result, tests run much
    quicker and have no issues with altered DB state after all tests ran
    when `--keepdb` is used.

    As a side effect, this test case does not support web browser automation.
    Use Django's testing client instead to test requests and responses.

    Also, it prevents the regular tests from running.
    """



"""Behave BDD integration for Django"""

__author__ = 'Mitchel Cabuloy'
__maintainer__ = 'Mitchel Cabuloy, Peter Bittner'
__email__ = 'mixxorz@gmail.com'
__url__ = 'https://github.com/behave/behave-django'
__license__ = 'MIT'
__version__ = '1.3.0'



def fixtures(*fixture_files):
    """
    Provide fixtures for given step_impl. Fixtures will be loaded in
    environment.py#before_scenario.

    @fixtures('data.json')
    @when('a user clicks the button')
    def step_impl(context):
        pass

    :param fixture_files: list of fixture files
    """

    def wrapper(step_impl):
        setattr(step_impl, 'registered_fixtures', fixture_files)
        return step_impl

    return wrapper



from __future__ import absolute_import
import sys

from behave.configuration import options as behave_options
from behave.__main__ import main as behave_main
from django.core.management.base import BaseCommand

from behave_django.environment import monkey_patch_behave
from behave_django.runner import (BehaviorDrivenTestRunner,
                                  ExistingDatabaseTestRunner,
                                  SimpleTestRunner)


def add_command_arguments(parser):
    """
    Additional command line arguments for the behave management command
    """
    parser.add_argument(
        '--noinput',
        '--no-input',
        action='store_const',
        const=False,
        dest='interactive',
        help='Tells Django to NOT prompt the user for input of any kind.',
    )
    parser.add_argument(
        '--failfast', action='store_const', const=True, dest='failfast',
        help=('Tells Django to stop running the '
              'test suite after first failed test.'),
    )
    parser.add_argument(
        '-r', '--reverse', action='store_const', const=True, dest='reverse',
        help='Reverses test cases order.',
    )
    parser.add_argument(
        '--use-existing-database',
        action='store_true',
        default=False,
        help="Don't create a test database. USE AT YOUR OWN RISK!",
    )
    parser.add_argument(
        '-k', '--keepdb',
        action='store_const',
        const=True,
        help="Preserves the test DB between runs.",
    )
    parser.add_argument(
        '-S', '--simple',
        action='store_true',
        default=False,
        help="Use simple test runner that supports Django's"
        " testing client only (no web browser automation)"
    )


def add_behave_arguments(parser):  # noqa
    """
    Additional command line arguments extracted directly from behave
    """

    # Option strings that conflict with Django
    conflicts = [
        '--no-color',
        '--version',
        '-c',
        '-k',
        '-v',
        '-S',
        '--simple',
    ]

    parser.add_argument(
        'paths',
        action='store',
        nargs='*',
        help="Feature directory, file or file location (FILE:LINE)."
    )

    for fixed, keywords in behave_options:
        keywords = keywords.copy()

        # Configfile only entries are ignored
        if not fixed:
            continue

        # Build option strings
        option_strings = []
        for option in fixed:
            # Prefix conflicting option strings with `--behave`
            if option in conflicts:
                prefix = '--' if option.startswith('--') else '-'
                option = option.replace(prefix, '--behave-', 1)

            option_strings.append(option)

        # config_help isn't a valid keyword for add_argument
        if 'config_help' in keywords:
            keywords['help'] = keywords['config_help']
            del keywords['config_help']

        parser.add_argument(*option_strings, **keywords)


class Command(BaseCommand):
    help = 'Runs behave tests'

    def add_arguments(self, parser):
        """
        Add behave's and our command line arguments to the command
        """
        parser.usage = "%(prog)s [options] [ [DIR|FILE|FILE:LINE] ]+"
        parser.description = """\
        Run a number of feature tests with behave."""

        add_command_arguments(parser)
        add_behave_arguments(parser)

    def handle(self, *args, **options):

        # Check the flags
        if options['use_existing_database'] and options['simple']:
            self.stderr.write(self.style.WARNING(
                '--simple flag has no effect'
                ' together with --use-existing-database'
            ))

        # Configure django environment
        passthru_args = ('failfast',
                         'interactive',
                         'keepdb',
                         'reverse')
        runner_args = {k: v for
                       k, v in
                       options.items() if k in passthru_args and v is not None}

        if options['dry_run'] or options['use_existing_database']:
            django_test_runner = ExistingDatabaseTestRunner(**runner_args)
        elif options['simple']:
            django_test_runner = SimpleTestRunner(**runner_args)
        else:
            django_test_runner = BehaviorDrivenTestRunner(**runner_args)

        django_test_runner.setup_test_environment()

        old_config = django_test_runner.setup_databases()

        # Run Behave tests
        monkey_patch_behave(django_test_runner)
        behave_args = self.get_behave_args()
        exit_status = behave_main(args=behave_args)

        # Teardown django environment
        django_test_runner.teardown_databases(old_config)
        django_test_runner.teardown_test_environment()

        if exit_status != 0:
            sys.exit(exit_status)

    def get_behave_args(self, argv=sys.argv):
        """
        Get a list of those command line arguments specified with the
        management command that are meant as arguments for running behave.
        """
        parser = BehaveArgsHelper().create_parser('manage.py', 'behave')
        args, unknown = parser.parse_known_args(argv[2:])

        behave_args = []
        for option in unknown:
            # Remove behave prefix
            if option.startswith('--behave-'):
                option = option.replace('--behave-', '', 1)
                prefix = '-' if len(option) == 1 else '--'
                option = prefix + option

            behave_args.append(option)

        return behave_args


class BehaveArgsHelper(Command):

    def add_arguments(self, parser):
        """
        Override setup of command line arguments to make behave commands not
        be recognized. The unrecognized args will then be for behave! :)
        """
        add_command_arguments(parser)



from copy import copy

from behave import step_registry as module_step_registry
from behave.runner import ModelRunner, Context
from django.shortcuts import resolve_url


class PatchedContext(Context):

    @property
    def base_url(self):
        try:
            return self.test.live_server_url
        except AttributeError:
            raise RuntimeError('Web browser automation is not available. '
                               'This scenario step can not be run with the '
                               '--simple or -S flag.')

    def get_url(self, to=None, *args, **kwargs):
        return self.base_url + (
            resolve_url(to, *args, **kwargs) if to else '')


def load_registered_fixtures(context):
    """
    Apply fixtures that are registered with the @fixtures decorator.
    """
    # -- SELECT STEP REGISTRY:
    # HINT: Newer behave versions use runner.step_registry
    # to be able to support multiple runners, each with its own step_registry.
    runner = context._runner    # pylint: disable=protected-access
    step_registry = getattr(runner, 'step_registry', None)
    if not step_registry:
        # -- BACKWARD-COMPATIBLE: Use module_step_registry
        step_registry = module_step_registry.registry

    # -- SETUP SCENARIO FIXTURES:
    for step in context.scenario.all_steps:
        match = step_registry.find_match(step)
        if match and hasattr(match.func, 'registered_fixtures'):
            if not context.test.fixtures:
                context.test.fixtures = []
            context.test.fixtures.extend(match.func.registered_fixtures)


class BehaveHooksMixin(object):
    """
    Provides methods that run during test execution

    These methods are attached to behave via monkey patching.
    """
    testcase_class = None

    def patch_context(self, context):
        """
        Patches the context to add utility functions

        Sets up the base_url, and the get_url() utility function.
        """
        context.__class__ = PatchedContext
        # Simply setting __class__ directly doesn't work
        # because behave.runner.Context.__setattr__ is implemented wrongly.
        object.__setattr__(context, '__class__', PatchedContext)

    def setup_testclass(self, context):
        """
        Adds the test instance to context
        """
        context.test = self.testcase_class()

    def setup_fixtures(self, context):
        """
        Sets up fixtures
        """
        if getattr(context, 'fixtures', None):
            context.test.fixtures = copy(context.fixtures)

        if getattr(context, 'reset_sequences', None):
            context.test.reset_sequences = context.reset_sequences

        if getattr(context, 'multi_db', None):
            context.test.__class__.multi_db = context.multi_db

        if hasattr(context, 'scenario'):
            load_registered_fixtures(context)

    def setup_test(self, context):
        """
        Sets up the Django test

        This method runs the code necessary to create the test database, start
        the live server, etc.
        """
        context.test._pre_setup(run=True)
        context.test.setUpClass()
        context.test()

    def teardown_test(self, context):
        """
        Tears down the Django test
        """
        context.test.tearDownClass()
        context.test._post_teardown(run=True)
        del context.test


def monkey_patch_behave(django_test_runner):
    """
    Integrate behave_django in behave via before/after scenario hooks
    """
    behave_run_hook = ModelRunner.run_hook

    def run_hook(self, name, context, *args):
        if name == 'before_all':
            django_test_runner.patch_context(context)

        behave_run_hook(self, name, context, *args)

        if name == 'before_scenario':
            django_test_runner.setup_testclass(context)
            django_test_runner.setup_fixtures(context)
            django_test_runner.setup_test(context)
            behave_run_hook(self, 'django_ready', context)

        if name == 'after_scenario':
            django_test_runner.teardown_test(context)

    ModelRunner.run_hook = run_hook



#!/usr/bin/env python3
"""
Packaging setup of Django integration for behave.
"""
from os import chdir
from os.path import abspath, dirname, normpath
from setuptools import find_packages, setup

# allow setup.py to be run from any path
chdir(normpath(abspath(dirname(__file__))))

import behave_django as package  # noqa


def read_file(filename):
    with open(filename) as file:
        return file.read()


setup(
    name='behave-django',
    version=package.__version__,
    license=package.__license__,
    description=package.__doc__.strip(),
    long_description=read_file('README.rst'),
    long_description_content_type='text/x-rst',
    packages=find_packages(exclude=['test*']),
    include_package_data=True,
    url=package.__url__,
    author=package.__author__,
    author_email=package.__email__,
    maintainer=package.__maintainer__,
    maintainer_email=package.__email__,
    install_requires=read_file('requirements.txt'),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: Plugins',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Testing',
    ],
    test_suite='tests',
)
